# import argparse
# import gym
# import random
# import Wolfpack_gym
# from Agent import LSTMAgent
# from torch.utils.tensorboard import SummaryWriter
# import numpy as np
# from gym
# from datetime import date
# import random
# import string
# import os
# import json
# import copy

import argparse
import gym
import random
import Wolfpack_gym
from Agent import DGNAgent
from torch.utils.tensorboard import SummaryWriter
import numpy as np
from gym.vector import AsyncVectorEnv
from stable_baselines.common.vec_env.subproc_vec_env import SubprocVecEnv
from datetime import date
import random
import string
import os
import json
import copy

parser = argparse.ArgumentParser()
parser.add_argument('--lr', type=float, default=1e-3, help='learning rate')
parser.add_argument('--gamma', type=float, default=0.99, help='dicount_rate')
parser.add_argument('--num_episodes', type=int, default=2000, help="Number of episodes for training")
parser.add_argument('--update_frequency', type=int, default=4, help="Timesteps between updates")
parser.add_argument('--saving_frequency', type=int,default=50,help="saving frequency")
parser.add_argument('--with_gpu', type=bool,default=False,help="with gpu")
parser.add_argument('--num_envs', type=int,default=16, help="Number of environments")
parser.add_argument('--tau', type=float,default=0.001, help="tau")
parser.add_argument('--clip_grad', type=float,default=10.0, help="gradient clipping")
parser.add_argument('--eval_eps', type=int, default=5, help="Evaluation episodes")
parser.add_argument('--weight_predict', type=float, default=1.0, help="Evaluation episodes")
parser.add_argument('--save_dir', type=str, default='parameters', help="parameter dir name")
parser.add_argument('--num_players', type=int, default=3, help="num players")
parser.add_argument('--close_penalty', type=float, default=0.1, help="close penalty")
parser.add_argument('--seed', type=int, default=0, help="seed")

args = parser.parse_args()

if __name__ == '__main__':

    args = vars(args)


    def make_env(env_id, rank, num_players, seed=1285, close_penalty=0.5, implicit_max_player_num=3,
                 with_shuffling=True, rnn_with_gnn=True, collapsed=True):
        def _init():
            env = gym.make(env_id, seed=seed + rank, num_players=num_players, close_penalty=close_penalty,
                           implicit_max_player_num=implicit_max_player_num, with_shuffling=with_shuffling,
                           rnn_with_gnn=rnn_with_gnn, collapsed=collapsed)
            return env

        return _init

    today = date.today()
    d1 = today.strftime("%d_%m_%Y")

    def randomString(stringLength=10):
        letters = string.ascii_lowercase
        return ''.join(random.choice(letters) for i in range(stringLength))

    random_experiment_name = randomString(10)
    writer = SummaryWriter(log_dir="runs/" + random_experiment_name)
    directory = os.path.join(args['save_dir'], random_experiment_name)
    if not os.path.exists(directory):
        os.makedirs(directory)

    with open(os.path.join(directory, 'params.json'), 'w') as json_file:
        json.dump(args, json_file)

    agent = DGNAgent(args=args)
    num_players = args['num_players']
    env = AsyncVectorEnv([make_env('Adhoc-wolfpack-v5', i, num_players,
                                   args['seed']) for i in range(args['num_envs'])])

    save_dirs = os.path.join(directory, 'params_0')
    agent.save_parameters(save_dirs)

    avgs = []
    for ep_val_num in range(args['eval_eps']):
        num_players = args['num_players']
        agent.reset(None)
        steps = 0
        avg_total_rewards = 0.0
        env_eval = AsyncVectorEnv([make_env('Adhoc-wolfpack-v5', i, num_players,
                                            2000, args['close_penalty']) for i in range(args['num_envs'])])
    
        f_done = False
        obs = env_eval.reset()
        steps = 0
        while not f_done:
            acts = agent.step(obs, eval=True)
            n_obs, rewards, dones, info = env_eval.step(acts)
            avg_total_rewards += (sum(rewards) + 0.0) / len(rewards)
            f_done = any(dones)
            obs = n_obs
            steps += 1
        avgs.append(avg_total_rewards)
        print("Finished eval with rewards " + str(avg_total_rewards))
    
    env_eval.close()
    writer.add_scalar('Rewards/train_set', sum(avgs) / len(avgs), 0)
    
    avgs = []
    for ep_val_num in range(args['eval_eps']):
        num_players = args['num_players']
        agent.reset(None)
        steps = 0
        avg_total_rewards = 0.0
        env_eval = AsyncVectorEnv([make_env('Adhoc-wolfpack-v5', i, num_players,
                                            2000, args['close_penalty'], implicit_max_player_num=5) for i in
                                   range(args['num_envs'])])
    
        f_done = False
        obs = env_eval.reset()
    
        while not f_done:
            acts = agent.step(obs, eval=True)
            n_obs, rewards, dones, info = env_eval.step(acts)
            avg_total_rewards += (sum(rewards) + 0.0) / len(rewards)
            f_done = any(dones)
            obs = n_obs
        avgs.append(avg_total_rewards)
        print("Finished eval with rewards " + str(avg_total_rewards))
    env_eval.close()
    writer.add_scalar('Rewards/eval', sum(avgs) / len(avgs), 0)

    for ep_num in range(args['num_episodes']):
        print(ep_num)
        train_avgs = 0
        agent.reset(None)
        agent.set_epsilon(max(1.0-((ep_num+0.0)/1500)*0.95,0.05))
        steps = 0

        f_done = False
        obs = env.reset()
        agent.compute_target(None, None, None, None, obs, add_storage=False)
        while not f_done:
            acts = agent.step(obs)
            n_obs, rewards, dones, info = env.step(acts)
            f_done = any(dones)

            n_obs_replaced = n_obs
            if f_done:
                n_obs_replaced = copy.deepcopy(n_obs)
                for idx in range(args['num_envs']):
                    if dones[idx]:
                        n_obs_replaced[idx] = info[idx]['terminal_observation']
            steps += 1

            train_avgs += (sum(rewards) + 0.0) / len(rewards)

            agent.compute_target(obs, rewards, acts, dones, n_obs_replaced, add_storage=True)
            if steps % args['update_frequency'] == 0 or f_done:
                agent.update()

            obs = n_obs

        writer.add_scalar('Rewards/train', train_avgs, ep_num)

        if (ep_num + 1) % args['saving_frequency'] == 0:
            save_dirs = os.path.join(directory, 'params_'+str((ep_num +
                                                               1) // args['saving_frequency']))
            agent.save_parameters(save_dirs)
            avgs = []
            for ep_val_num in range(args['eval_eps']):
                num_players = args['num_players']
                agent.reset(None)
                steps = 0
                avg_total_rewards = 0.0
                env_eval = AsyncVectorEnv([make_env('Adhoc-wolfpack-v5', i, num_players,
                                                    2000, args['close_penalty']) for i in range(args['num_envs'])])

                f_done = False
                obs = env_eval.reset()

                while not f_done:
                    acts = agent.step(obs, eval=True)
                    n_obs, rewards, dones, info = env_eval.step(acts)
                    avg_total_rewards += (sum(rewards) + 0.0) / len(rewards)
                    f_done = any(dones)
                    obs = n_obs
                avgs.append(avg_total_rewards)
                print("Finished eval with rewards " + str(avg_total_rewards))
            env_eval.close()
            writer.add_scalar('Rewards/train_set', sum(avgs) / len(avgs),
                              (ep_num + 1) // args['saving_frequency'])

            avgs = []
            for ep_val_num in range(args['eval_eps']):
                num_players = args['num_players']
                agent.reset(None)
                steps = 0
                avg_total_rewards = 0.0
                env_eval = AsyncVectorEnv([make_env('Adhoc-wolfpack-v5', i, num_players,
                                                    2000, args['close_penalty'], implicit_max_player_num=5) for i in
                                           range(args['num_envs'])])

                f_done = False
                obs = env_eval.reset()

                while not f_done:
                    acts = agent.step(obs, eval=True)
                    n_obs, rewards, dones, info = env_eval.step(acts)
                    avg_total_rewards += (sum(rewards) + 0.0) / len(rewards)
                    f_done = any(dones)
                    obs = n_obs
                avgs.append(avg_total_rewards)
                print("Finished eval with rewards " + str(avg_total_rewards))
            env_eval.close()
            writer.add_scalar('Rewards/eval', sum(avgs) / len(avgs),
                              (ep_num + 1) // args['saving_frequency'])
